@@ -1,6 +1,50 @@
 from functools import wraps
 from json import dumps, loads
 from base64 import b64encode
+from inspect import signature
+
+def compact_dump(value):
+    return dumps(value, separators=(',', ':'))
+
+def get_args(fn, args, kwargs):
+    """
+    This function parses the args and kwargs in the context of a function and creates unified
+    dictionary of {<argument_name>: <value>}. This is useful
+    because arguments can be passed as args or kwargs, and we want to make sure we cache
+    them both the same. Otherwise there would be different caching for add(1, 2) and add(arg1=1, arg2=2)
+    """
+    arg_sig = signature(fn)
+    standard_args = [param.name for param in arg_sig.parameters.values() if param.kind is param.POSITIONAL_OR_KEYWORD]
+    allowed_kwargs = {param.name for param in arg_sig.parameters.values() if param.kind is param.POSITIONAL_OR_KEYWORD or param.kind is param.KEYWORD_ONLY}
+    variable_args = [param.name for param in arg_sig.parameters.values() if param.kind is param.VAR_POSITIONAL]
+    variable_kwargs = [param.name for param in arg_sig.parameters.values() if param.kind is param.VAR_KEYWORD]
+    parsed_args = {}
+
+    if standard_args or variable_args:
+        for index, arg in enumerate(args):
+            try:
+                parsed_args[standard_args[index]] = arg
+            except IndexError:
+                # then fallback to using the positional varargs name
+                if variable_args:
+                    vargs_name = variable_args[0]
+                    if vargs_name not in parsed_args:
+                        parsed_args[vargs_name] = []
+
+                    parsed_args[vargs_name].append(arg)
+
+
+    if kwargs:
+        for key, value in kwargs.items():
+            if key in allowed_kwargs:
+                parsed_args[key] = value
+            elif variable_kwargs:
+                vkwargs_name = variable_kwargs[0]
+                if vkwargs_name not in parsed_args:
+                    parsed_args[vkwargs_name] = {}
+                parsed_args[vkwargs_name][key] = value
+
+    return parsed_args
 
 
 def get_cache_lua_fn(client):
@@ -55,21 +99,25 @@ def chunks(iterable, n):
 
 
 class RedisCache:
-    def __init__(self, redis_client, prefix="rc", serializer=dumps, deserializer=loads):
+    def __init__(self, redis_client, prefix="rc", serializer=compact_dump, deserializer=loads, key_serializer=None, exception_handler=None):
         self.client = redis_client
         self.prefix = prefix
         self.serializer = serializer
         self.deserializer = deserializer
+        self.key_serializer = key_serializer
+        self.exception_handler = exception_handler
 
-    def cache(self, ttl=0, limit=0, namespace=None):
+    def cache(self, ttl=0, limit=0, namespace=None, exception_handler=None):
         return CacheDecorator(
             redis_client=self.client,
             prefix=self.prefix,
             serializer=self.serializer,
             deserializer=self.deserializer,
+            key_serializer=self.key_serializer,
             ttl=ttl,
             limit=limit,
-            namespace=namespace
+            namespace=namespace,
+            exception_handler=exception_handler or self.exception_handler
         )
 
     def mget(self, *fns_with_args):
@@ -105,43 +153,70 @@ class RedisCache:
         return deserialized_results
 
 class CacheDecorator:
-    def __init__(self, redis_client, prefix="rc", serializer=dumps, deserializer=loads, ttl=0, limit=0, namespace=None):
+    def __init__(self, redis_client, prefix="rc", serializer=compact_dump, deserializer=loads, key_serializer=None, ttl=0, limit=0, namespace=None, exception_handler=None):
         self.client = redis_client
         self.prefix = prefix
         self.serializer = serializer
+        self.key_serializer = key_serializer
         self.deserializer = deserializer
         self.ttl = ttl
         self.limit = limit
         self.namespace = namespace
+        self.exception_handler = exception_handler
         self.keys_key = None
+        self.original_fn = None
+
+
+    def get_full_prefix(self):
+        return f'{{{self.prefix}:{self.namespace}}}'
 
     def get_key(self, args, kwargs):
-        serialized_data = self.serializer([args, kwargs])
+        normalized_args = get_args(self.original_fn, args, kwargs)
 
-        if not isinstance(serialized_data, str):
-            serialized_data = str(b64encode(serialized_data), 'utf-8')
-        return f'{self.prefix}:{self.namespace}:{serialized_data}'
+        if self.key_serializer:
+            serialized_data = self.key_serializer(normalized_args)
+        else:
+            serialized_data = self.serializer(normalized_args)
+
+        if isinstance(serialized_data, str):
+            serialized_data = serialized_data.encode('utf-8')
+
+        # Encode the value as base64 to avoid issues with {} and other special characters
+        serialized_encoded_data = b64encode(serialized_data).decode('utf-8')
+
+        return f'{self.get_full_prefix()}:{serialized_encoded_data}'
 
     def __call__(self, fn):
-        self.namespace = self.namespace if self.namespace else f'{fn.__module__}.{fn.__name__}'
-        self.keys_key = f'{self.prefix}:{self.namespace}:keys'
+        self.namespace = self.namespace or f'{fn.__module__}.{fn.__qualname__}'
+        self.keys_key = f'{self.get_full_prefix()}:keys'
         self.original_fn = fn
 
         @wraps(fn)
         def inner(*args, **kwargs):
             nonlocal self
             key = self.get_key(args, kwargs)
-            result = self.client.get(key)
-            if not result:
-                result = fn(*args, **kwargs)
-                result_serialized = self.serializer(result)
+            result = None
+
+            exception_handled = False
+            try:
+                result = self.client.get(key)
+            except Exception as e:
+                if self.exception_handler:
+                    # This allows people to handle failures in cache lookups
+                    exception_handled = True
+                    parsed_result = self.exception_handler(e, self.original_fn, args, kwargs)
+            if result:
+                parsed_result = self.deserializer(result)
+            elif not exception_handled:
+                parsed_result = fn(*args, **kwargs)
+                result_serialized = self.serializer(parsed_result)
                 get_cache_lua_fn(self.client)(keys=[key, self.keys_key], args=[result_serialized, self.ttl, self.limit])
-            else:
-                result = self.deserializer(result)
-            return result
+
+            return parsed_result
 
         inner.invalidate = self.invalidate
         inner.invalidate_all = self.invalidate_all
+        inner.get_full_prefix = self.get_full_prefix
         inner.instance = self
         return inner
 
@@ -153,6 +228,6 @@ class CacheDecorator:
         pipe.execute()
 
     def invalidate_all(self, *args, **kwargs):
-        chunks_gen = chunks(self.client.scan_iter(f'{self.prefix}:{self.namespace}:*'), 500)
+        chunks_gen = chunks(self.client.scan_iter(f'{self.get_full_prefix()}:*'), 500)
         for keys in chunks_gen:
             self.client.delete(*keys)
